// Compressor
//------------------------------------------------------------------------------

// Includes
//------------------------------------------------------------------------------
#include "Compressor.h"

#include "Core/Containers/AutoPtr.h"
#include "Core/Env/Assert.h"
#include "Core/Env/Types.h"
#include "Core/Math/Conversions.h"
#include "Core/Mem/Mem.h"

#include <memory.h>

// Static Data & Typedefs
//------------------------------------------------------------------------------
typedef uint32_t AccumulatorType;
const size_t AccumulatorSizeBits( sizeof( AccumulatorType ) * 8 );

//------------------------------------------------------------------------------
Compressor::Compressor()
	: m_Result( nullptr )
	, m_ResultSize( 0 )
{
}

// DESTRUCTOR
//------------------------------------------------------------------------------
Compressor::~Compressor()
{
	FREE( m_Result );
}

// IsValidData
//------------------------------------------------------------------------------
bool Compressor::IsValidData( const void * data, size_t dataSize ) const
{
	const Header * header = (const Header *)data;
	if ( header->m_CompressionType > 1 )
	{
		return false;
	}
	if ( ( header->m_CompressedSize + sizeof( Header ) ) != dataSize )
	{
		return false;
	}
	if ( header->m_CompressedSize > header->m_UncompressedSize )
	{
		return false;
	}
	return true;
}

// Compress
//------------------------------------------------------------------------------
#pragma warning( push )
#pragma warning( disable : 6262 ) // static analysis warns about stack usage of predictionTable (64KB)
bool Compressor::Compress( const void * data, size_t dataSize )
{
	ASSERT( data );
	ASSERT( ( (uintptr_t)data % 4 ) == 0 ); // input must be 4 byte aligned
	ASSERT( m_Result == nullptr );

	// create and initialize prediction table
	unsigned char predictionTable[ 256 * 256 ];
	memset( predictionTable, 0, 256 * 256 );

	// allocate space for worst case ( mispredict 1 bit cost for every byte ) 
	size_t worstCaseSpace = dataSize +			// input data
							( dataSize / 8 ) +	// with no compression is a 1 bit per byte overhead
							1 +					// plus rounding
							sizeof( AccumulatorType );// rounding

	AutoPtr< char > workspace( (char *)ALLOC( worstCaseSpace ) );

	// compressed data
	AccumulatorType * compressedData = (AccumulatorType *)workspace.Get();
	memset( compressedData, 0, worstCaseSpace );

	// iterate through data
	const unsigned char * pos = (const unsigned char *)data;
	const unsigned char * const end = pos + dataSize;

	size_t posInBits = 0;
	size_t predictionIndex( 0 );

	do
	{
		const unsigned char c = *pos++;
		const unsigned char predictedC = predictionTable[ predictionIndex ];
		predictionTable[ predictionIndex ] = c;

		size_t bitsNum = ( c == predictedC ) ? 1 : 9;
		size_t bitMask = ( c == predictedC ) ? 0 : 0x1FF; // 1 zeroed bit / 9 bits
		size_t bitsVal = ( bitMask & ( ( c << 1 ) | 0x1 ) );

		size_t outPos = ( (size_t)compressedData + ( posInBits >> 3 ) & ~1);
		*( unsigned int *)outPos |= bitsVal << ( posInBits & 0xF );

		posInBits += bitsNum;

		predictionIndex <<= 8;
		predictionIndex += c;
		predictionIndex &= 0xFFFF;
	}
	while( pos < end );

	uint32_t compressedSize = uint32_t( posInBits >> 3 );
	compressedSize += ( ( posInBits - ( (size_t)compressedSize << 3 ) ) > 0 ) ? 1 : 0;
	compressedSize = Math::RoundUp( compressedSize, 4 );
	ASSERT( compressedSize <= worstCaseSpace );

	// did the compression yield any benefit?
	const bool compressed = ( compressedSize < dataSize );

	if ( compressed )
	{
		// trim memory usage to compressed size
		m_Result = ALLOC( compressedSize + sizeof( Header ) );
		memcpy( (char *)m_Result + sizeof( Header ), compressedData, compressedSize );
		m_ResultSize = compressedSize + sizeof( Header );
	}
	else
	{
		// compression failed, so just copy the old data
		m_Result = ALLOC( dataSize + sizeof( Header ) );
		memcpy( (char *)m_Result + sizeof( Header ), data, dataSize );
		m_ResultSize = dataSize + sizeof( Header );
	}

	// fill out header
	Header * header = (Header*)m_Result;
	header->m_CompressionType = compressed ? 1 : 0;		// compression type
	header->m_UncompressedSize = (uint32_t)dataSize;	// input size
	header->m_CompressedSize = compressed ? compressedSize : (uint32_t)dataSize;	// output size

	return compressed;
}
#pragma warning( pop )

// Decompress
//------------------------------------------------------------------------------
#pragma warning( push )
#pragma warning( disable : 6262 ) // static analysis warns about stack usage of predictionTable (64KB)
void Compressor::Decompress( const void * data )
{
	ASSERT( data );
	ASSERT( ( (uintptr_t)data % 4 ) == 0 ); // output must be 4 byte aligned
	ASSERT( m_Result == nullptr );

	const Header * header = (const Header *)data;

	// handle uncompressed case
	if ( header->m_CompressionType == 0 )
	{
		m_Result = ALLOC( header->m_UncompressedSize );
		memcpy( m_Result, (char *)data + sizeof( Header ), header->m_UncompressedSize );
		m_ResultSize = header->m_UncompressedSize;
		return;
	}
	ASSERT( header->m_CompressionType == 1 );

	// uncompressed size
	const uint32_t uncompressedSize = header->m_UncompressedSize;

	// first 4 bytes contains decompression size
	AccumulatorType * compressedData = (AccumulatorType *)( (intptr_t)data + sizeof( Header ) );

	// allocate space for output
	m_Result = (char *)ALLOC( uncompressedSize );

	// create and initialize prediction table
	unsigned char predictionTable[ 256 * 256 ];
	memset( predictionTable, 0, 256 * 256 );

	unsigned char * uncompressedData = (unsigned char *)m_Result;
	const unsigned char * const uncompressedDataEnd( uncompressedData + uncompressedSize );

	size_t posInBits = 0;
	size_t predictionIndex( 0 );

	do
	{		
		// read 9 bits
		size_t inPos = ( (size_t)compressedData + ( posInBits >> 3 ) & ~1);
		size_t bits = *( unsigned int *)inPos;
		bits >>= ( posInBits & 0xF );
		size_t isMiss = ( bits & 0x1 );
		posInBits += 1 + ( 8 * isMiss );
		size_t charBits = ( bits >> 1 );

		unsigned char predictedC = predictionTable[ predictionIndex ];
		size_t keepMask		= ( isMiss * 0xFF );
		size_t keepMaskInv	= 0xFF & ~keepMask;
		unsigned char c = (unsigned char)( ( keepMask & charBits ) | ( keepMaskInv & predictedC ) );
		predictionTable[ predictionIndex ] = c;
				
		*uncompressedData++ = c;

		predictionIndex <<= 8;
		predictionIndex += c;
		predictionIndex &= 0xFFFF;
	}
	while ( uncompressedData != uncompressedDataEnd );

	m_ResultSize = uncompressedSize;
}
#pragma warning( pop )

//------------------------------------------------------------------------------
